
import unittest
import json

def json_proc(value):
    return json.loads(json.dumps(value))

class FuzzyFloat(float):
    def __new__(self, value, places=7):
        del places
        return float.__new__(self, value)

    def __init__(self, value, places=7):
        float.__init__(value)
        self.__places = places
    def __eq__(self, other):
        if isinstance(other, float):
            return 2 * abs(self - other) <= 10 ** (-self.__places)
        else:
            return super().__eq__(self, other)
    def __repr__(self):
        return repr(round(self, self.__places + 2))

def fuzzify(obj, places):
    if isinstance(obj, str):
        return obj
    if isinstance(obj, list):
        return [fuzzify(x, places) for x in obj]
    if isinstance(obj, dict):
        return {a : fuzzify(b, places) for a, b in obj.items()}
    if isinstance(obj, (float, int)):
        return FuzzyFloat(obj)
    raise RuntimeError("Invalid object type for fuzzification: {}".format(type(obj)))

class TestCase2(unittest.TestCase):
    def assertContainersAlmostEqual(self, first, second, places=7, canonicalizer=json_proc, **kwargs):
        """
        Checks if two container items are almost equal. Numbers are compared via `self.assertAlmostEqual`.

        Arguments:
            first, second: things to check
            canonicalizer: thing that converts the containers to only json-compatible types
        """
        first = fuzzify(canonicalizer(first), places)
        second = fuzzify(canonicalizer(second), places)
        self.assertEqual(first, second, **kwargs)




from .canvas import Canvas

def _forward(method):
    def func(self, *args, **kwargs):
        # pylint: disable=W0212
        return getattr(self._canvas, method)(*args, **kwargs)
    func.name = method
    return func

class ForwardingCanvas(Canvas):
    """
    Canvas that dispatches all calls to a contained canvas
    """
    def __init__(self, canvas):
        super().__init__(-1, -1)
        self._canvas = canvas

    def set_canvas(self, canvas):
        canvas.turtle = self._canvas.turtle
        self._canvas = canvas

    draw_rectangular_line = _forward("draw_rectangular_line")
    draw_circle = _forward("draw_circle")
    fill_path = _forward("fill_path")
    axis_aligned_rectangle = _forward("axis_aligned_rectangle")
    set_bgcolor = _forward("set_bgcolor")
    clear = _forward("clear")
    refreshed_turtle = _forward("refreshed_turtle")



COLORS = {
  "aliceblue": "#f0f8ff",
  "antiquewhite": "#faebd7",
  "aqua": "#00ffff",
  "aquamarine": "#7fffd4",
  "azure": "#f0ffff",
  "beige": "#f5f5dc",
  "bisque": "#ffe4c4",
  "black": "#000000",
  "blanchedalmond": "#ffebcd",
  "blue": "#0000ff",
  "blueviolet": "#8a2be2",
  "brown": "#a52a2a",
  "burlywood": "#deb887",
  "cadetblue": "#5f9ea0",
  "chartreuse": "#7fff00",
  "chocolate": "#d2691e",
  "coral": "#ff7f50",
  "cornflowerblue": "#6495ed",
  "cornsilk": "#fff8dc",
  "crimson": "#dc143c",
  "cyan": "#00ffff",
  "darkblue": "#00008b",
  "darkcyan": "#008b8b",
  "darkgoldenrod": "#b8860b",
  "darkgray": "#a9a9a9",
  "darkgreen": "#006400",
  "darkkhaki": "#bdb76b",
  "darkmagenta": "#8b008b",
  "darkolivegreen": "#556b2f",
  "darkorange": "#ff8c00",
  "darkorchid": "#9932cc",
  "darkred": "#8b0000",
  "darksalmon": "#e9967a",
  "darkseagreen": "#8fbc8f",
  "darkslateblue": "#483d8b",
  "darkslategray": "#2f4f4f",
  "darkturquoise": "#00ced1",
  "darkviolet": "#9400d3",
  "deeppink": "#ff1493",
  "deepskyblue": "#00bfff",
  "dimgray": "#696969",
  "dodgerblue": "#1e90ff",
  "firebrick": "#b22222",
  "floralwhite": "#fffaf0",
  "forestgreen": "#228b22",
  "fuchsia": "#ff00ff",
  "gainsboro": "#dcdcdc",
  "ghostwhite": "#f8f8ff",
  "gold": "#ffd700",
  "goldenrod": "#daa520",
  "gray": "#808080",
  "green": "#008000",
  "greenyellow": "#adff2f",
  "honeydew": "#f0fff0",
  "hotpink": "#ff69b4",
  "indianred ": "#cd5c5c",
  "indigo": "#4b0082",
  "ivory": "#fffff0",
  "khaki": "#f0e68c",
  "lavender": "#e6e6fa",
  "lavenderblush": "#fff0f5",
  "lawngreen": "#7cfc00",
  "lemonchiffon": "#fffacd",
  "lightblue": "#add8e6",
  "lightcoral": "#f08080",
  "lightcyan": "#e0ffff",
  "lightgoldenrodyellow": "#fafad2",
  "lightgrey": "#d3d3d3",
  "lightgreen": "#90ee90",
  "lightpink": "#ffb6c1",
  "lightsalmon": "#ffa07a",
  "lightseagreen": "#20b2aa",
  "lightskyblue": "#87cefa",
  "lightslategray": "#778899",
  "lightsteelblue": "#b0c4de",
  "lightyellow": "#ffffe0",
  "lime": "#00ff00",
  "limegreen": "#32cd32",
  "linen": "#faf0e6",
  "magenta": "#ff00ff",
  "maroon": "#800000",
  "mediumaquamarine": "#66cdaa",
  "mediumblue": "#0000cd",
  "mediumorchid": "#ba55d3",
  "mediumpurple": "#9370d8",
  "mediumseagreen": "#3cb371",
  "mediumslateblue": "#7b68ee",
  "mediumspringgreen": "#00fa9a",
  "mediumturquoise": "#48d1cc",
  "mediumvioletred": "#c71585",
  "midnightblue": "#191970",
  "mintcream": "#f5fffa",
  "mistyrose": "#ffe4e1",
  "moccasin": "#ffe4b5",
  "navajowhite": "#ffdead",
  "navy": "#000080",
  "oldlace": "#fdf5e6",
  "olive": "#808000",
  "olivedrab": "#6b8e23",
  "orange": "#ffa500",
  "orangered": "#ff4500",
  "orchid": "#da70d6",
  "palegoldenrod": "#eee8aa",
  "palegreen": "#98fb98",
  "paleturquoise": "#afeeee",
  "palevioletred": "#d87093",
  "papayawhip": "#ffefd5",
  "peachpuff": "#ffdab9",
  "peru": "#cd853f",
  "pink": "#ffc0cb",
  "plum": "#dda0dd",
  "powderblue": "#b0e0e6",
  "purple": "#800080",
  "rebeccapurple": "#663399",
  "red": "#ff0000",
  "rosybrown": "#bc8f8f",
  "royalblue": "#4169e1",
  "saddlebrown": "#8b4513",
  "salmon": "#fa8072",
  "sandybrown": "#f4a460",
  "seagreen": "#2e8b57",
  "seashell": "#fff5ee",
  "sienna": "#a0522d",
  "silver": "#c0c0c0",
  "skyblue": "#87ceeb",
  "slateblue": "#6a5acd",
  "slategray": "#708090",
  "snow": "#fffafa",
  "springgreen": "#00ff7f",
  "steelblue": "#4682b4",
  "tan": "#d2b48c",
  "teal": "#008080",
  "thistle": "#d8bfd8",
  "tomato": "#ff6347",
  "turquoise": "#40e0d0",
  "violet": "#ee82ee",
  "wheat": "#f5deb3",
  "white": "#ffffff",
  "whitesmoke": "#f5f5f5",
  "yellow": "#ffff00",
  "yellowgreen": "#9acd32"
}




import unittest

from collections import Counter

from abstract_turtle import LoggingCanvas
from abstract_turtle.turtle import *

class TestTurtleModule(unittest.TestCase):

    def test_module_import_star(self):
        canvas = LoggingCanvas(1000, 1000)
        set_canvas(canvas)
        fd(100)
        lt(90)
        self.assertAlmostEqual(100, xcor())
        self.assertAlmostEqual(0, ycor())
        self.assertAlmostEqual(90, heading())
        self.assertEqual(
            {'draw_rectangular_line' : 1, 'refreshed_turtle' : 3},
            Counter([func for func, *_ in canvas.log])
        )




from math import pi

from collections import Counter

from abstract_turtle import Turtle, LoggingCanvas

from .utils import TestCase2

class BasicSquareTest(TestCase2):

    def setUp(self, width=1000, height=1000):
        self.canvas = LoggingCanvas(width, height)
        self.turtle = Turtle(self.canvas)

    def assertPositionAlmostEqual(self, t, pos):
        self.assertContainersAlmostEqual(t.pos(), pos)

    def test_square(self):
        t = self.turtle
        self.assertPositionAlmostEqual(t, (0, 0))
        t.fd(100)
        t.rt(90)
        self.assertPositionAlmostEqual(t, (100, 0))
        t.fd(100)
        t.rt(90)
        self.assertPositionAlmostEqual(t, (100, -100))
        t.fd(100)
        t.rt(90)
        self.assertPositionAlmostEqual(t, (0, -100))
        t.fd(100)
        t.rt(90)
        self.assertPositionAlmostEqual(t, (0, 0))

        self.assertEqual(
            {'draw_rectangular_line' : 4, 'refreshed_turtle' : 9},
            Counter([func for func, *_ in self.canvas.log])
        )

    def test_degrees(self):
        t = self.turtle
        t.rt(45)
        self.assertEqual(360-45, t.heading())
        t.degrees(400)
        self.assertAlmostEqual(400-50, t.heading())
        t.radians()
        self.assertAlmostEqual(2*pi - pi/4, t.heading())

    def test_set_heading(self):
        t = self.turtle
        t.setheading(45)
        self.assertEqual(45, t.heading())

    def test_home(self):
        t = self.turtle
        t.rt(45)
        t.fd(100)
        t.home()
        self.assertAlmostEqual(0, t.heading())
        self.assertPositionAlmostEqual(t, (0, 0))

    def test_pensize(self):
        t = self.turtle
        self.assertEqual(1, t.pensize())
        t.pensize(10)
        self.assertEqual(10, t.pensize())

    def test_penupdown(self):
        t = self.turtle
        self.assertEqual(True, t.isdown())
        t.penup()
        self.assertEqual(False, t.isdown())

    def test_full_circle(self):
        t = self.turtle
        t.circle(2)
        self.assertContainersAlmostEqual(
            self.canvas.log,
            [
                ['refreshed_turtle', [[0, 0], 0, 1, 1]],
                ['draw_circle', [0, 2, 2], [0, 0, 0], 1, False, -pi/2, 3*pi/2],
                ['refreshed_turtle', [[0, 0], 0, 1, 1]],
            ]
        )

    def test_part_circle(self):
        t = self.turtle
        t.circle(2, 45)
        self.assertContainersAlmostEqual(
            self.canvas.log,
            [
                ['refreshed_turtle', [[0, 0], 0, 1, 1]],
                ['draw_circle', [0, 2, 2], [0, 0, 0], 1, False, -pi/2, -pi/4],
                ['refreshed_turtle', [[2**0.5, 2 - 2**0.5], pi/4, 1, 1]],
            ]
        )

    def test_rotated_full_circle(self):
        t = self.turtle
        t.lt(45)
        t.circle(2, 45)
        self.assertContainersAlmostEqual(
            self.canvas.log,
            [
                ['refreshed_turtle', [[0, 0], 0, 1, 1]],
                ['refreshed_turtle', [[0, 0], pi/4, 1, 1]],
                ['draw_circle', [-2**0.5, 2**0.5, 2], [0, 0, 0], 1, False, -pi/4, 0],
                ['refreshed_turtle', [[2 - 2**0.5, 2**0.5], pi/2, 1, 1]],
            ]
        )

    def test_partial_circle_fill(self):
        t = self.turtle
        t.pu()
        t.begin_fill()
        t.fillcolor("red")
        t.fd(100)
        t.circle(100, 90)
        t.fd(100)
        t.end_fill()
        self.assertContainersAlmostEqual(
            self.canvas.log,
            [
                ['refreshed_turtle', [[0, 0],       0, 1, 1]],
                ['refreshed_turtle', [[100, 0],     0, 1, 1]],
                ['refreshed_turtle', [[200, 100],  pi/2, 1, 1]],
                ['refreshed_turtle', [[200, 200],  pi/2, 1, 1]],
                ['fill_path', [
                        ['line', [0, 0]],
                        ['line', [100, 0]],
                        ['arc', [100, 100], 100, -pi / 2, 0],
                        ['line', [200, 200]]
                    ],
                    [255, 0, 0]
                ]
            ]
        )

    def test_distance(self):
        t1 = self.turtle
        t1.setpos(3, 8)
        t2 = Turtle(LoggingCanvas(1000, 1000))
        t2.setpos(6, 12)

        self.assertAlmostEqual(t1.distance(t2), 5)
        self.assertAlmostEqual(t1.distance((0, 5)), 3 * 2**0.5)

    def test_bad_color_string(self):
        t1 = self.turtle
        try:
            t1.bgcolor('reds')
            self.fail()
        except RuntimeError as e:
            self.assertEqual(str(e), "Invalid color string: 'reds'")

    def test_bad_hex_color(self):
        t1 = self.turtle
        try:
            t1.bgcolor('#hello')
            self.fail()
        except RuntimeError as e:
            self.assertEqual(str(e), "Invalid hex color string: 'hello'")

    def test_bad_hex_color_parsing(self):
        t1 = self.turtle
        try:
            t1.bgcolor('#bcdefg')
            self.fail()
        except RuntimeError as e:
            self.assertEqual(str(e), "Invalid hex color string: 'bcdefg'")

    def test_bad_integer_color(self):
        t1 = self.turtle
        try:
            t1.bgcolor(0, 280, 30)
            self.fail()
        except RuntimeError as e:
            self.assertEqual(str(e), "Invalid integer color: (0, 280, 30)")

    def test_bad_types(self):
        t1 = self.turtle
        try:
            t1.bgcolor(0, 40.0, 30)
            self.fail()
        except RuntimeError as e:
            self.assertEqual(str(e), "Invalid color. Expected either 3 ints or 1 string, but got: int, float, int")

    def test_mode(self):
        t = self.turtle
        t.mode("standard")
        self.assertEqual(t.mode(), "standard")
        t.mode("logo")
        self.assertEqual(t.mode(), "logo")

    def test_width_height_default(self):
        t = self.turtle
        self.assertEqual(t.canvas_width(), 1000)
        self.assertEqual(t.canvas_height(), 1000)
        t.pixel_size(2)
        self.assertEqual(t.canvas_width(), 500)
        self.assertEqual(t.canvas_height(), 500)
        t.pixel_size(3)
        self.assertEqual(t.canvas_width(), 333)
        self.assertEqual(t.canvas_height(), 333)

    def test_width_height_diff_sizes(self):
        self.setUp(200, 500)
        t = self.turtle
        self.assertEqual(t.canvas_width(), 200)
        self.assertEqual(t.canvas_height(), 500)
        t.pixel_size(2)
        self.assertEqual(t.canvas_width(), 100)
        self.assertEqual(t.canvas_height(), 250)
        t.pixel_size(3)
        self.assertEqual(t.canvas_width(), 66)
        self.assertEqual(t.canvas_height(), 166)



from abc import ABC, abstractmethod
from collections import namedtuple
from enum import Enum

from math import sin, cos

from .color_names import COLORS


class Color(namedtuple('Color', ['red', 'green', 'blue'])):
    @staticmethod
    def of(*color):
        if len(color) == 3 and all(isinstance(c, int) for c in color):
            if not all(0 <= c < 256 for c in color):
                raise RuntimeError("Invalid integer color: {!r}".format(color))
            return Color(*color)
        if len(color) == 1 and isinstance(color[0], str):
            color = color[0].lower()
            if color and color[0] == "#":
                return Color.hexcolor(color[1:])
            if color in COLORS:
                return Color.of(COLORS[color])
            raise RuntimeError("Invalid color string: {!r}".format(color))
        types = [type(c).__name__ for c in color]
        raise RuntimeError("Invalid color. Expected either 3 ints or 1 string, but got: {}".format(", ".join(types)))

    @staticmethod
    def hexcolor(color):
        if len(color) == 3:  # shorthand hex
            color = "".join(x * 2 for x in color)
        if len(color) == 6:
            vals = [Color.hexparse(color[i:i+2]) for i in range(0, 6, 2)]
            if all(x is not None for x in vals):
                return Color.of(*vals)
        raise RuntimeError("Invalid hex color string: {!r}".format(color))

    @staticmethod
    def hexparse(pair):
        try:
            return int(pair, 16)
        except ValueError:
            return None


Position = namedtuple('Position', ['x', 'y'])

class DrawnTurtle(namedtuple('DrawnTurtle', ['pos', 'heading', 'stretch_wid', 'stretch_len'])):
    @property
    def points(self):
        unadjusted_points = [
            (-3, 8),
            (0, 0),
            (-3, -8),
            (8, 0),
        ]
        stretched_points = [
            (dx * self.stretch_len, dy * self.stretch_wid) for dx, dy in unadjusted_points
        ]
        rotated_points = [
            rotate(*dxy, self.heading) for dxy in stretched_points
        ]
        moved_points = [
            Position(self.pos.x + dx, self.pos.y + dy) for dx, dy in rotated_points
        ]
        return moved_points

    @property
    def json_friendly(self):
        return [
            [self.pos.x, self.pos.y],
            self.heading,
            self.stretch_wid,
            self.stretch_len
        ]


class Mode(Enum):
    LOGO = "logo"
    STANDARD = "standard"


def rotate(x, y, theta):
    return x * cos(theta) - y * sin(theta), x * sin(theta) + y * cos(theta)


class Path(ABC):
    @abstractmethod
    def to_points(self, divisions=100):
        """
        Return an iterable of points corresponding to this path element
        """

    @property
    @abstractmethod
    def json_friendly(self):
        """
        Converts the given path to a JSON compatible object.
        """


class LineTo(Path, namedtuple("LineTo", ["position"])):
    def to_points(self, divisions=100):
        del divisions
        return [self.position]

    @property
    def json_friendly(self):
        return ["line", self.position]


class Arc(Path, namedtuple("Arc", ["center", "radius", "start_angle", "end_angle"])):
    """
    CENTER is the center of the arc
    START_ANGLE and END_ANGLE are the starting and ending angles of the arc measured in radians CCW from the x-axis
    """

    def to_points(self, divisions=100):
        for i in range(divisions + 1):
            ang = self.start_angle + (i / divisions) * (self.end_angle - self.start_angle)
            yield Position(self.center.x + self.radius * cos(ang), self.center.y + self.radius * sin(ang))

    @property
    def json_friendly(self):
        return ["arc", self.center, self.radius, self.start_angle, self.end_angle]




from math import sin, cos, pi
try:
    from PIL import Image, ImageDraw
    import numpy as np
except ImportError:
    pass

from .canvas import Canvas
from .model import Color, Position


class PillowCanvas(Canvas):
    """
    Draw the given results on a pillow canvas.
    """
    def __init__(self, width, height):
        super().__init__(width, height)
        self.image = Image.new('RGBA', (width, height))
        self.background_color = Color(255, 255, 255)
        self.draw = ImageDraw.Draw(self.image)
        self.turtle = None

    def tr_pos(self, pos):
        x, y = pos.x, pos.y
        return np.round(x + self.width / 2), np.round(-y + self.height / 2)

    @staticmethod
    def tr_color(color):
        return color.red, color.green, color.blue, 255

    def draw_rectangular_line(self, start, end, color, width):
        self.draw.line([self.tr_pos(start), self.tr_pos(end)], self.tr_color(color), width)

    def draw_circle(self, center, radius, color, width, is_filled, start, end):
        x, y = self.tr_pos(center)
        left_up = (x-radius, y-radius)
        right_down = (x+radius, y+radius)
        box = [left_up, right_down]
        if is_filled:
            assert start == 0
            assert end == 2 * pi
            self.draw.ellipse(box, fill=self.tr_color(color))
        else:
            self.draw.arc(box, -end * 180 / pi, -start * 180 / pi, fill=self.tr_color(color), width=width)

    def fill_path(self, path, color):
        points = []
        for movement in path:
            points += movement.to_points()
        self.draw.polygon(
            [self.tr_pos(point) for point in points],
            fill=self.tr_color(color)
        )

    def axis_aligned_rectangle(self, bottom_left, width, height, color):
        blx, bly = self.tr_pos(bottom_left)
        # bounding box in pixel space, make this exactly w*h
        # subtract one to have blx, blx+1, ..., blx+w-1, etc.
        tlx, tly = blx + (width - 1), bly - (height - 1)
        bounding_box = [blx, bly, tlx, tly]
        print(bounding_box)
        self.draw.rectangle(
            bounding_box,
            fill=self.tr_color(color)
        )

    def set_bgcolor(self, color):
        self.background_color = color

    def clear(self):
        self.draw.rectangle((0, 0, self.width, self.height), fill=(0, 0, 0, 0))

    def export(self):
        image = self.image.copy()
        if self.turtle is not None:
            ImageDraw.Draw(image).polygon(
                [self.tr_pos(point) for point in self.turtle.points],
                fill=self.tr_color(Color.of(0, 0, 0))
            )
        data = np.array(image)
        assert len(data.shape) == 3 and data.shape[-1] == 4
        transparents = data[:,:,-1] == 0
        data[transparents] = self.tr_color(self.background_color)
        return Image.fromarray(data)

    def refreshed_turtle(self, turtle):
        # no need to do stuff
        pass
